@@ -10,25 +10,21 @@ from typing import Pattern
 from .exceptions import JSONPathSyntaxError
 from .token import TOKEN_AND
 from .token import TOKEN_BARE_PROPERTY
-from .token import TOKEN_BRACKET_PROPERTY
 from .token import TOKEN_COMMA
 from .token import TOKEN_CONTAINS
 from .token import TOKEN_DDOT
-from .token import TOKEN_DOT_INDEX
 from .token import TOKEN_DOT_PROPERTY
 from .token import TOKEN_DOUBLE_QUOTE_STRING
 from .token import TOKEN_EQ
 from .token import TOKEN_FALSE
+from .token import TOKEN_FILTER
 from .token import TOKEN_FILTER_CONTEXT
-from .token import TOKEN_FILTER_END
-from .token import TOKEN_FILTER_START
 from .token import TOKEN_FLOAT
 from .token import TOKEN_FUNCTION
 from .token import TOKEN_GE
 from .token import TOKEN_GT
 from .token import TOKEN_ILLEGAL
 from .token import TOKEN_IN
-from .token import TOKEN_INDEX
 from .token import TOKEN_INT
 from .token import TOKEN_INTERSECTION
 from .token import TOKEN_KEY
@@ -56,11 +52,9 @@ from .token import TOKEN_RPAREN
 from .token import TOKEN_SELF
 from .token import TOKEN_SINGLE_QUOTE_STRING
 from .token import TOKEN_SKIP
-from .token import TOKEN_SLICE
 from .token import TOKEN_SLICE_START
 from .token import TOKEN_SLICE_STEP
 from .token import TOKEN_SLICE_STOP
-from .token import TOKEN_STRING
 from .token import TOKEN_TRUE
 from .token import TOKEN_UNDEFINED
 from .token import TOKEN_UNION
@@ -85,28 +79,12 @@ class Lexer:
         # .thing
         self.dot_property_pattern = rf"\.(?P<G_PROP>{self.key_pattern})"
 
-        # [thing]
-        self.bracketed_property_pattern = rf"\[\s*(?P<G_BPROP>{self.key_pattern})\s*]"
-
-        # [1] or [-1]
-        self.index_pattern = r"\[\s*(?P<G_INDEX>\-?\s*\d+)\s*]"
-
-        # [:] or [1:-1] or [1:] or [:1] or [-1:] or [:-1] or [::] or [-1:0:-1]
-        self.slice_pattern = (
-            r"\[\s*(?P<G_SLICE_START>\-?\d*)\s*"
-            r":\s*(?P<G_SLICE_STOP>\-?\d*)\s*"
-            r"(?::\s*(?P<G_SLICE_STEP>\-?\d*))?\s*]"
-        )
-
         self.slice_list_pattern = (
             r"(?P<G_LSLICE_START>\-?\d*)\s*"
             r":\s*(?P<G_LSLICE_STOP>\-?\d*)\s*"
             r"(?::\s*(?P<G_LSLICE_STEP>\-?\d*))?"
         )
 
-        # .* or [*] or .[*]
-        self.wild_pattern = r"\.?(?:\[\s*\*\s*]|\*)"
-
         # `not` or !
         self.logical_not_pattern = r"(?:not|!)"
 
@@ -130,17 +108,11 @@ class Lexer:
             (TOKEN_DOUBLE_QUOTE_STRING, self.double_quote_pattern),
             (TOKEN_SINGLE_QUOTE_STRING, self.single_quote_pattern),
             (TOKEN_RE_PATTERN, self.re_pattern),
-            (TOKEN_INDEX, self.index_pattern),
-            (TOKEN_SLICE, self.slice_pattern),
-            (TOKEN_WILD, self.wild_pattern),
             (TOKEN_LIST_SLICE, self.slice_list_pattern),
-            (TOKEN_FILTER_START, r"\[\s*\?\s*\(?"),
-            (TOKEN_FILTER_END, r"\)\s*]"),
             (TOKEN_FUNCTION, self.function_pattern),
-            (TOKEN_BRACKET_PROPERTY, self.bracketed_property_pattern),
             (TOKEN_DOT_PROPERTY, self.dot_property_pattern),
             (TOKEN_FLOAT, r"-?\d+\.\d*(?:e[+-]?\d+)?"),
-            (TOKEN_INT, r"-?\d+(?:e[+\-]?\d+)?\b"),
+            (TOKEN_INT, r"-?\d+(?P<G_EXP>e[+\-]?\d+)?\b"),
             (TOKEN_DDOT, r"\.\."),
             (TOKEN_AND, self.bool_and_pattern),
             (TOKEN_OR, self.bool_or_pattern),
@@ -151,6 +123,8 @@ class Lexer:
             (TOKEN_INTERSECTION, re.escape(self.env.intersection_token)),
             (TOKEN_FILTER_CONTEXT, re.escape(self.env.filter_context_token)),
             (TOKEN_KEYS, re.escape(self.env.keys_selector_token)),
+            (TOKEN_WILD, r"\*"),
+            (TOKEN_FILTER, r"\?"),
             (TOKEN_IN, r"in"),
             (TOKEN_TRUE, r"[Tt]rue"),
             (TOKEN_FALSE, r"[Ff]alse"),
@@ -198,12 +172,6 @@ class Lexer:
                     value=match.group("G_PROP"),
                     index=match.start("G_PROP"),
                 )
-            elif kind == TOKEN_BRACKET_PROPERTY:
-                yield _token(
-                    kind=TOKEN_PROPERTY,
-                    value=match.group("G_BPROP"),
-                    index=match.start("G_BPROP"),
-                )
             elif kind == TOKEN_BARE_PROPERTY:
                 yield _token(
                     kind=TOKEN_BARE_PROPERTY,
@@ -226,46 +194,31 @@ class Lexer:
                     value=match.group("G_LSLICE_STEP") or "",
                     index=match.start("G_LSLICE_STEP"),
                 )
-            elif kind == TOKEN_DOT_INDEX:
-                yield _token(
-                    kind=TOKEN_INDEX,
-                    value=match.group("G_DINDEX"),
-                    index=match.start("G_DINDEX"),
-                )
-            elif kind == TOKEN_INDEX:
-                yield _token(
-                    kind=TOKEN_INDEX,
-                    value=match.group("G_INDEX"),
-                    index=match.start("G_INDEX"),
-                )
-            elif kind == TOKEN_SLICE:
-                yield _token(
-                    kind=TOKEN_SLICE_START,
-                    value=match.group("G_SLICE_START"),
-                    index=match.start("G_SLICE_START"),
-                )
-                yield _token(
-                    kind=TOKEN_SLICE_STOP,
-                    value=match.group("G_SLICE_STOP"),
-                    index=match.start("G_SLICE_STOP"),
-                )
-                yield _token(
-                    kind=TOKEN_SLICE_STEP,
-                    value=match.group("G_SLICE_STEP") or "",
-                    index=match.start("G_SLICE_STEP"),
-                )
             elif kind == TOKEN_DOUBLE_QUOTE_STRING:
                 yield _token(
-                    kind=TOKEN_STRING,
+                    kind=TOKEN_DOUBLE_QUOTE_STRING,
                     value=match.group("G_DQUOTE"),
                     index=match.start("G_DQUOTE"),
                 )
             elif kind == TOKEN_SINGLE_QUOTE_STRING:
                 yield _token(
-                    kind=TOKEN_STRING,
+                    kind=TOKEN_SINGLE_QUOTE_STRING,
                     value=match.group("G_SQUOTE"),
                     index=match.start("G_SQUOTE"),
                 )
+            elif kind == TOKEN_INT:
+                if match.group("G_EXP") and match.group("G_EXP")[1] == "-":
+                    yield _token(
+                        kind=TOKEN_FLOAT,
+                        value=match.group(),
+                        index=match.start(),
+                    )
+                else:
+                    yield _token(
+                        kind=TOKEN_INT,
+                        value=match.group(),
+                        index=match.start(),
+                    )
             elif kind == TOKEN_RE_PATTERN:
                 yield _token(
                     kind=TOKEN_RE_PATTERN,
