@@ -1,7 +1,7 @@
 """The default JSONPath parser."""
 from __future__ import annotations
 
-import codecs
+import json
 import re
 from typing import TYPE_CHECKING
 from typing import Callable
@@ -11,7 +11,11 @@ from typing import List
 from typing import Optional
 from typing import Union
 
+from jsonpath.function_extensions.filter_function import ExpressionType
+from jsonpath.function_extensions.filter_function import FilterFunction
+
 from .exceptions import JSONPathSyntaxError
+from .exceptions import JSONPathTypeError
 from .filter import CURRENT_KEY
 from .filter import FALSE
 from .filter import NIL
@@ -25,6 +29,7 @@ from .filter import FunctionExtension
 from .filter import InfixExpression
 from .filter import IntegerLiteral
 from .filter import ListLiteral
+from .filter import Path
 from .filter import PrefixExpression
 from .filter import RegexLiteral
 from .filter import RootPath
@@ -45,18 +50,17 @@ from .token import TOKEN_BARE_PROPERTY
 from .token import TOKEN_COMMA
 from .token import TOKEN_CONTAINS
 from .token import TOKEN_DDOT
+from .token import TOKEN_DOUBLE_QUOTE_STRING
 from .token import TOKEN_EOF
 from .token import TOKEN_EQ
 from .token import TOKEN_FALSE
+from .token import TOKEN_FILTER
 from .token import TOKEN_FILTER_CONTEXT
-from .token import TOKEN_FILTER_END
-from .token import TOKEN_FILTER_START
 from .token import TOKEN_FLOAT
 from .token import TOKEN_FUNCTION
 from .token import TOKEN_GE
 from .token import TOKEN_GT
 from .token import TOKEN_IN
-from .token import TOKEN_INDEX
 from .token import TOKEN_INT
 from .token import TOKEN_INTERSECTION
 from .token import TOKEN_KEY
@@ -81,14 +85,15 @@ from .token import TOKEN_RE_PATTERN
 from .token import TOKEN_ROOT
 from .token import TOKEN_RPAREN
 from .token import TOKEN_SELF
+from .token import TOKEN_SINGLE_QUOTE_STRING
 from .token import TOKEN_SLICE_START
 from .token import TOKEN_SLICE_STEP
 from .token import TOKEN_SLICE_STOP
-from .token import TOKEN_STRING
 from .token import TOKEN_TRUE
 from .token import TOKEN_UNDEFINED
 from .token import TOKEN_UNION
 from .token import TOKEN_WILD
+from .token import Token
 
 if TYPE_CHECKING:
     from .env import JSONPathEnvironment
@@ -175,6 +180,17 @@ class Parser:
         TOKEN_RE: "=~",
     }
 
+    SINGULAR_QUERY_COMPARISON_OPERATORS = frozenset(
+        [
+            "==",
+            ">=",
+            ">",
+            "<=",
+            "<",
+            "!=",
+        ]
+    )
+
     PREFIX_OPERATORS = frozenset(
         [
             TOKEN_NOT,
@@ -212,7 +228,8 @@ class Parser:
             TOKEN_ROOT: self.parse_root_path,
             TOKEN_SELF: self.parse_self_path,
             TOKEN_FILTER_CONTEXT: self.parse_filter_context_path,
-            TOKEN_STRING: self.parse_string_literal,
+            TOKEN_DOUBLE_QUOTE_STRING: self.parse_string_literal,
+            TOKEN_SINGLE_QUOTE_STRING: self.parse_string_literal,
             TOKEN_TRUE: self.parse_boolean,
             TOKEN_UNDEFINED: self.parse_undefined,
             TOKEN_FUNCTION: self.parse_function_extension,
@@ -225,7 +242,8 @@ class Parser:
             TOKEN_NIL: self.parse_nil,
             TOKEN_NONE: self.parse_nil,
             TOKEN_NULL: self.parse_nil,
-            TOKEN_STRING: self.parse_string_literal,
+            TOKEN_DOUBLE_QUOTE_STRING: self.parse_string_literal,
+            TOKEN_SINGLE_QUOTE_STRING: self.parse_string_literal,
             TOKEN_TRUE: self.parse_boolean,
         }
 
@@ -239,11 +257,13 @@ class Parser:
             TOKEN_NIL: self.parse_nil,
             TOKEN_NONE: self.parse_nil,
             TOKEN_NULL: self.parse_nil,
-            TOKEN_STRING: self.parse_string_literal,
+            TOKEN_SINGLE_QUOTE_STRING: self.parse_string_literal,
+            TOKEN_DOUBLE_QUOTE_STRING: self.parse_string_literal,
             TOKEN_TRUE: self.parse_boolean,
             TOKEN_ROOT: self.parse_root_path,
             TOKEN_SELF: self.parse_self_path,
             TOKEN_FILTER_CONTEXT: self.parse_filter_context_path,
+            TOKEN_FUNCTION: self.parse_function_extension,
         }
 
     def parse(self, stream: TokenStream) -> Iterable[JSONPathSelector]:
@@ -271,20 +291,21 @@ class Parser:
                     env=self.env,
                     token=stream.current,
                     name=stream.current.value,
+                    shorthand=True,
                 )
-            elif stream.current.kind == TOKEN_INDEX:
-                yield self.parse_index(stream)
             elif stream.current.kind == TOKEN_SLICE_START:
                 yield self.parse_slice(stream)
             elif stream.current.kind == TOKEN_WILD:
                 yield WildSelector(
                     env=self.env,
                     token=stream.current,
+                    shorthand=True,
                 )
             elif stream.current.kind == TOKEN_KEYS:
                 yield KeysSelector(
                     env=self.env,
                     token=stream.current,
+                    shorthand=True,
                 )
             elif stream.current.kind == TOKEN_DDOT:
                 yield RecursiveDescentSelector(
@@ -293,8 +314,6 @@ class Parser:
                 )
             elif stream.current.kind == TOKEN_LIST_START:
                 yield self.parse_selector_list(stream)
-            elif stream.current.kind == TOKEN_FILTER_START:
-                yield self.parse_filter(stream)
             else:
                 if in_filter:
                     stream.push(stream.current)
@@ -302,20 +321,6 @@ class Parser:
 
             stream.next_token()
 
-    def parse_index(self, stream: TokenStream) -> IndexSelector:
-        """Parse an index selector from a stream of tokens."""
-        if (
-            len(stream.current.value) > 1 and stream.current.value.startswith("0")
-        ) or stream.current.value.startswith("-0"):
-            raise JSONPathSyntaxError(
-                "leading zero in index selector", token=stream.current
-            )
-        return IndexSelector(
-            env=self.env,
-            token=stream.current,
-            index=int(stream.current.value),
-        )
-
     def parse_slice(self, stream: TokenStream) -> SliceSelector:
         """Parse a slice JSONPath expression from a stream of tokens."""
         start_token = stream.next_token()
@@ -347,7 +352,7 @@ class Parser:
             step=step,
         )
 
-    def parse_selector_list(self, stream: TokenStream) -> ListSelector:
+    def parse_selector_list(self, stream: TokenStream) -> ListSelector:  # noqa: PLR0912
         """Parse a comma separated list JSONPath selectors from a stream of tokens."""
         tok = stream.next_token()
         list_items: List[
@@ -357,11 +362,19 @@ class Parser:
                 PropertySelector,
                 SliceSelector,
                 WildSelector,
+                Filter,
             ]
         ] = []
 
         while stream.current.kind != TOKEN_RBRACKET:
             if stream.current.kind == TOKEN_INT:
+                if (
+                    len(stream.current.value) > 1
+                    and stream.current.value.startswith("0")
+                ) or stream.current.value.startswith("-0"):
+                    raise JSONPathSyntaxError(
+                        "leading zero in index selector", token=stream.current
+                    )
                 list_items.append(
                     IndexSelector(
                         env=self.env,
@@ -375,6 +388,7 @@ class Parser:
                         env=self.env,
                         token=stream.current,
                         name=stream.current.value,
+                        shorthand=False,
                     ),
                 )
             elif stream.current.kind == TOKEN_KEYS:
@@ -382,34 +396,48 @@ class Parser:
                     KeysSelector(
                         env=self.env,
                         token=stream.current,
+                        shorthand=False,
                     )
                 )
-            elif stream.current.kind == TOKEN_STRING:
+            elif stream.current.kind in (
+                TOKEN_DOUBLE_QUOTE_STRING,
+                TOKEN_SINGLE_QUOTE_STRING,
+            ):
                 if self.RE_INVALID_NAME_SELECTOR.search(stream.current.value):
                     raise JSONPathSyntaxError(
                         f"invalid name selector {stream.current.value!r}",
                         token=stream.current,
                     )
 
-                name = (
-                    codecs.decode(
-                        stream.current.value.replace("\\/", "/"), "unicode-escape"
-                    )
-                    .encode("utf-16", "surrogatepass")
-                    .decode("utf-16")
-                )
-
                 list_items.append(
                     PropertySelector(
                         env=self.env,
                         token=stream.current,
-                        name=name,
+                        name=self._decode_string_literal(stream.current),
+                        shorthand=False,
                     ),
                 )
             elif stream.current.kind == TOKEN_SLICE_START:
                 list_items.append(self.parse_slice(stream))
             elif stream.current.kind == TOKEN_WILD:
-                list_items.append(WildSelector(env=self.env, token=stream.current))
+                list_items.append(
+                    WildSelector(
+                        env=self.env,
+                        token=stream.current,
+                        shorthand=False,
+                    )
+                )
+            elif stream.current.kind == TOKEN_FILTER:
+                list_items.append(self.parse_filter(stream))
+            elif stream.current.kind == TOKEN_EOF:
+                raise JSONPathSyntaxError(
+                    "unexpected end of query", token=stream.current
+                )
+            else:
+                raise JSONPathSyntaxError(
+                    f"unexpected token in bracketed selection {stream.current.kind!r}",
+                    token=stream.current,
+                )
 
             if stream.peek.kind == TOKEN_EOF:
                 raise JSONPathSyntaxError(
@@ -424,20 +452,26 @@ class Parser:
             stream.next_token()
 
         if not list_items:
-            raise JSONPathSyntaxError("empty segment", token=tok)
+            raise JSONPathSyntaxError("empty bracketed segment", token=tok)
 
         return ListSelector(env=self.env, token=tok, items=list_items)
 
     def parse_filter(self, stream: TokenStream) -> Filter:
         tok = stream.next_token()
-        expr = BooleanExpression(self.parse_filter_selector(stream))
+        expr = self.parse_filter_selector(stream)
 
-        if stream.peek.kind == TOKEN_RPAREN:
-            raise JSONPathSyntaxError("unbalanced ')'", token=stream.current)
+        if self.env.well_typed and isinstance(expr, FunctionExtension):
+            func = self.env.function_extensions.get(expr.name)
+            if (
+                func
+                and isinstance(func, FilterFunction)
+                and func.return_type == ExpressionType.VALUE
+            ):
+                raise JSONPathTypeError(
+                    f"result of {expr.name}() must be compared", token=tok
+                )
 
-        stream.next_token()
-        stream.expect(TOKEN_FILTER_END, TOKEN_RBRACKET)
-        return Filter(env=self.env, token=tok, expression=expr)
+        return Filter(env=self.env, token=tok, expression=BooleanExpression(expr))
 
     def parse_boolean(self, stream: TokenStream) -> FilterExpression:
         if stream.current.kind == TOKEN_TRUE:
@@ -451,10 +485,11 @@ class Parser:
         return UNDEFINED_LITERAL
 
     def parse_string_literal(self, stream: TokenStream) -> FilterExpression:
-        return StringLiteral(value=stream.current.value)
+        return StringLiteral(value=self._decode_string_literal(stream.current))
 
     def parse_integer_literal(self, stream: TokenStream) -> FilterExpression:
-        return IntegerLiteral(value=int(stream.current.value))
+        # Convert to float first to handle scientific notation.
+        return IntegerLiteral(value=int(float(stream.current.value)))
 
     def parse_float_literal(self, stream: TokenStream) -> FilterExpression:
         return FloatLiteral(value=float(stream.current.value))
@@ -474,11 +509,17 @@ class Parser:
     ) -> FilterExpression:
         tok = stream.next_token()
         precedence = self.PRECEDENCES.get(tok.kind, self.PRECEDENCE_LOWEST)
-        return InfixExpression(
-            left,
-            self.BINARY_OPERATORS[tok.kind],
-            self.parse_filter_selector(stream, precedence),
-        )
+        right = self.parse_filter_selector(stream, precedence)
+        operator = self.BINARY_OPERATORS[tok.kind]
+
+        self._raise_for_non_singular_query(left, tok)  # TODO: store tok on expression
+        self._raise_for_non_singular_query(right, tok)
+
+        if operator in self.SINGULAR_QUERY_COMPARISON_OPERATORS:
+            self._raise_for_non_comparable_function(left, tok)
+            self._raise_for_non_comparable_function(right, tok)
+
+        return InfixExpression(left, operator, right)
 
     def parse_grouped_expression(self, stream: TokenStream) -> FilterExpression:
         stream.next_token()
@@ -490,14 +531,9 @@ class Parser:
                 raise JSONPathSyntaxError(
                     "unbalanced parentheses", token=stream.current
                 )
-            if stream.current.kind == TOKEN_FILTER_END:
-                # In some cases, an RPAREN followed by an RBRACKET can
-                # look like the long form "end of filter" token.
-                stream.push(stream.current)
-                break
             expr = self.parse_infix_expression(stream, expr)
 
-        stream.expect(TOKEN_RPAREN, TOKEN_FILTER_END)
+        stream.expect(TOKEN_RPAREN)
         return expr
 
     def parse_root_path(self, stream: TokenStream) -> FilterExpression:
@@ -557,27 +593,34 @@ class Parser:
 
         while stream.current.kind != TOKEN_RPAREN:
             try:
-                function_arguments.append(
-                    self.function_argument_map[stream.current.kind](stream)
-                )
+                func = self.function_argument_map[stream.current.kind]
             except KeyError as err:
                 raise JSONPathSyntaxError(
                     f"unexpected {stream.current.value!r}",
                     token=stream.current,
                 ) from err
 
+            expr = func(stream)
+
+            # The argument could be a comparison or logical expression
+            peek_kind = stream.peek.kind
+            while peek_kind in self.BINARY_OPERATORS:
+                stream.next_token()
+                expr = self.parse_infix_expression(stream, expr)
+                peek_kind = stream.peek.kind
+
+            function_arguments.append(expr)
+
             if stream.peek.kind != TOKEN_RPAREN:
-                if stream.peek.kind == TOKEN_FILTER_END:
-                    break
                 stream.expect_peek(TOKEN_COMMA)
                 stream.next_token()
 
             stream.next_token()
 
-        function_arguments = self.env.validate_function_extension_signature(
-            tok, function_arguments
+        return FunctionExtension(
+            tok.value,
+            self.env.validate_function_extension_signature(tok, function_arguments),
         )
-        return FunctionExtension(tok.value, function_arguments)
 
     def parse_filter_selector(
         self, stream: TokenStream, precedence: int = PRECEDENCE_LOWEST
@@ -585,7 +628,7 @@ class Parser:
         try:
             left = self.token_map[stream.current.kind](stream)
         except KeyError as err:
-            if stream.current.kind in (TOKEN_EOF, TOKEN_FILTER_END, TOKEN_RBRACKET):
+            if stream.current.kind in (TOKEN_EOF, TOKEN_RBRACKET):
                 msg = "end of expression"
             else:
                 msg = repr(stream.current.value)
@@ -596,7 +639,7 @@ class Parser:
         while True:
             peek_kind = stream.peek.kind
             if (
-                peek_kind in (TOKEN_EOF, TOKEN_FILTER_END, TOKEN_RBRACKET)
+                peek_kind in (TOKEN_EOF, TOKEN_RBRACKET)
                 or self.PRECEDENCES.get(peek_kind, self.PRECEDENCE_LOWEST) < precedence
             ):
                 break
@@ -608,3 +651,42 @@ class Parser:
             left = self.parse_infix_expression(stream, left)
 
         return left
+
+    def _decode_string_literal(self, token: Token) -> str:
+        if self.env.unicode_escape:
+            if token.kind == TOKEN_SINGLE_QUOTE_STRING:
+                value = token.value.replace('"', '\\"').replace("\\'", "'")
+            else:
+                value = token.value
+            try:
+                rv = json.loads(f'"{value}"')
+                assert isinstance(rv, str)
+                return rv
+            except json.JSONDecodeError as err:
+                raise JSONPathSyntaxError(str(err).split(":")[1], token=token) from None
+
+        return token.value
+
+    def _raise_for_non_singular_query(
+        self, expr: FilterExpression, token: Token
+    ) -> None:
+        if (
+            self.env.well_typed
+            and isinstance(expr, Path)
+            and not expr.path.singular_query()
+        ):
+            raise JSONPathSyntaxError(
+                "non-singular query is not comparable", token=token
+            )
+
+    def _raise_for_non_comparable_function(
+        self, expr: FilterExpression, token: Token
+    ) -> None:
+        if not self.env.well_typed or not isinstance(expr, FunctionExtension):
+            return
+        func = self.env.function_extensions.get(expr.name)
+        if (
+            isinstance(func, FilterFunction)
+            and func.return_type != ExpressionType.VALUE
+        ):
+            raise JSONPathTypeError(f"result of {expr.name}() is not comparable", token)
